"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const elasticsearch_1 = require("@elastic/elasticsearch");
const command_1 = require("@oclif/command");
const cli_ux_1 = require("cli-ux");
const fs = require("fs");
const loadYamlFile = require("load-yaml-file");
const _ = require("lodash");
const path = require("path");
const base_1 = require("../base");
const chunkArray_1 = require("../utils/misc/chunkArray");
class CfRepos extends base_1.default {
    /*
      The aim of this script is to update which repositories should be active or inactive.
      It does the following:
       - Fetch a list of repositories from Elasticsearch
       - Search for each repository in the configuration file, to find if it should be enabled or disabled
       - Send back the content to Elasticsearch
     */
    async run() {
        const { flags } = this.parse(CfRepos);
        const userConfig = await loadYamlFile(path.join(this.config.configDir, 'config.yml'));
        const { esport, eshost, esrepo } = flags;
        const es_port = (esport !== undefined ? esport : userConfig.elasticsearch.port);
        const es_host = (eshost !== undefined ? eshost : userConfig.elasticsearch.host);
        const reposIndexName = (esrepo !== undefined ? esrepo : userConfig.elasticsearch.indices.repos);
        //1- Test if an index exists, if it does not, exit.
        cli_ux_1.default.action.start('Checking if index: ' + reposIndexName + ' exists');
        const client = new elasticsearch_1.Client({ node: es_host + ':' + es_port });
        const healthCheck = await client.cluster.health();
        if (healthCheck.body.status === 'red') {
            this.log('Elasticsearch cluster is not in an healthy state, exiting');
            this.log(healthCheck.body);
            process.exit(1);
        }
        const testIndex = await client.indices.exists({ index: reposIndexName });
        if (testIndex.body === false) {
            this.log('Could not find the index, please run ghRepos first');
        }
        cli_ux_1.default.action.stop(' done');
        //2- Grab the repositories data from ElasticSearch
        cli_ux_1.default.action.start('Grabbing existing data from ElasticSearch');
        let esRepos = await client.search({
            index: reposIndexName,
            body: {
                from: 0,
                size: 10000,
                query: {
                    match_all: {}
                }
            }
        });
        cli_ux_1.default.action.stop(' done');
        //3- Load content from the configuration file
        cli_ux_1.default.action.start('Grabbing repo configuration from file: ' + path.join(this.config.configDir, 'repositories.yml'));
        let reposConfig = [];
        if (fs.existsSync(path.join(this.config.configDir, 'repositories.yml'))) {
            reposConfig = await loadYamlFile(path.join(this.config.configDir, 'repositories.yml'));
        }
        else {
            this.error('Unable to find the repositories config file (' + path.join(this.config.configDir, 'repositories.yml') + '), please run ghRepos first', { exit: 1 });
        }
        cli_ux_1.default.action.stop(' done');
        //4- Update the array with the value from the configuration file
        cli_ux_1.default.action.start('Comparing Elasticsearch data with flags in configuration file');
        const updatedData = esRepos.body.hits.hits.map(repo => {
            const cfgRepo = _.find(reposConfig, (o) => o[repo._source.org.login + '/' + repo._source.name] !== undefined);
            if (repo._source.active !== cfgRepo[repo._source.org.login + '/' + repo._source.name]) {
                this.log('Changing: ' + repo._source.org.login + '/' + repo._source.name + ' from: ' + repo._source.active + ' to: ' + cfgRepo[repo._source.org.login + '/' + repo._source.name]);
            }
            return Object.assign({}, repo._source, { active: cfgRepo[repo._source.org.login + '/' + repo._source.name] });
        });
        cli_ux_1.default.action.stop(' done');
        //5- Push the results back to Elastic Search in chunks of 100
        cli_ux_1.default.action.start('Pushing data back to Elasticsearch');
        //Split the array in chunks of 100
        const esPayloadChunked = await chunkArray_1.default(updatedData, 100);
        for (const [idx, esPayloadChunk] of esPayloadChunked.entries()) {
            cli_ux_1.default.action.start('Submitting data to ElasticSearch (' + (idx + 1) + ' / ' + esPayloadChunked.length + ')');
            let formattedData = '';
            for (let rec of esPayloadChunk) {
                formattedData = formattedData + JSON.stringify({
                    index: {
                        _index: reposIndexName,
                        _id: rec.id
                    }
                }) + '\n' + JSON.stringify(rec) + '\n';
            }
            await client.bulk({ index: reposIndexName, refresh: 'wait_for', body: formattedData });
            cli_ux_1.default.action.stop(' done');
        }
    }
}
CfRepos.description = 'Enable/disable repositories by reading the configuration file';
CfRepos.examples = [
    '$ github-indexer cfRepo',
];
CfRepos.flags = Object.assign({}, base_1.default.flags, { help: command_1.flags.help({ char: 'h' }) });
exports.default = CfRepos;
