"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const elasticsearch_1 = require("@elastic/elasticsearch");
const command_1 = require("@oclif/command");
const cli_ux_1 = require("cli-ux");
const jsYaml = require("js-yaml");
const loadYamlFile = require("load-yaml-file");
const path = require("path");
const base_1 = require("../base");
const milestones_1 = require("../schemas/milestones");
const settings_1 = require("../schemas/settings");
const index_1 = require("../utils/github/fetchMilestones/index");
const chunkArray_1 = require("../utils/misc/chunkArray");
class GhMilestones extends base_1.default {
    /*
      The aim of this script is to fetch all issues associated with active repositories.
      It does the following:
       - Fetch a list of repositories from Elasticsearch
       - Fetch updated issues for each repository
       - Send back the content to Elasticsearch
     */
    async run() {
        const { flags } = this.parse(GhMilestones);
        const userConfig = await loadYamlFile(path.join(this.config.configDir, 'config.yml'));
        const { esport, eshost, esrepo } = flags;
        const es_port = (esport !== undefined ? esport : userConfig.elasticsearch.port);
        const es_host = (eshost !== undefined ? eshost : userConfig.elasticsearch.host);
        const reposIndexName = (esrepo !== undefined ? esrepo : userConfig.elasticsearch.indices.repos);
        const indexMilestonePrefix = userConfig.elasticsearch.indices.milestones;
        //1- Test if an index exists, if it does not, create it.
        cli_ux_1.default.action.start('Checking if index: ' + reposIndexName + ' exists');
        const client = new elasticsearch_1.Client({ node: es_host + ':' + es_port });
        const healthCheck = await client.cluster.health();
        if (healthCheck.body.status === 'red') {
            this.log('Elasticsearch cluster is not in an healthy state, exiting');
            this.log(healthCheck.body);
            process.exit(1);
        }
        const testIndex = await client.indices.exists({ index: reposIndexName });
        if (testIndex.body === false) {
            this.error('Index: ' + reposIndexName + ' does not exists, please configure repositories (ghRepos) first', { exit: 1 });
        }
        cli_ux_1.default.action.stop(' done');
        //2- Grab the active repositories from Elasticsearch
        cli_ux_1.default.action.start('Grabbing the active repositories from ElasticSearch');
        let esRepos = await client.search({
            index: reposIndexName,
            body: {
                from: 0,
                size: 10000,
                query: {
                    match: {
                        active: {
                            query: true
                        }
                    }
                }
            }
        });
        //console.log(esRepos.body.hits.hits);
        const activeRepos = esRepos.body.hits.hits.map(r => r._source);
        cli_ux_1.default.action.stop(' done');
        if (activeRepos.length === 0) {
            this.error('The script could not find any active repositories. Please use ghRepos and cfRepos first.', { exit: 1 });
        }
        const fetchData = new index_1.default(this.log, userConfig, this.config.configDir, cli_ux_1.default);
        this.log('Starting to grab issues');
        for (let repo of activeRepos) {
            //A - Check if repo index exists, if not create
            const milestonesIndex = (indexMilestonePrefix + repo.org.login + '_' + repo.name).toLocaleLowerCase();
            const testIndex = await client.indices.exists({ index: milestonesIndex });
            if (testIndex.body === false) {
                cli_ux_1.default.action.start('Elasticsearch Index ' + milestonesIndex + ' does not exist, creating');
                const mappings = await jsYaml.safeLoad(milestones_1.default);
                const settings = await jsYaml.safeLoad(settings_1.default);
                await client.indices.create({ index: milestonesIndex, body: { settings, mappings } });
                cli_ux_1.default.action.stop(' created');
            }
            //B - Find the most recent issue
            let searchResult = await client.search({
                index: milestonesIndex,
                body: {
                    query: {
                        match_all: {}
                    },
                    size: 1,
                    sort: [
                        {
                            updatedAt: {
                                order: 'desc'
                            }
                        }
                    ]
                }
            });
            let recentMilestone = null;
            if (searchResult.body.hits.hits.length > 0) {
                recentMilestone = searchResult.body.hits.hits[0]._source;
            }
            //C - Fetch issues from GitHub into a large array
            cli_ux_1.default.action.start('Grabbing issues for: ' + repo.org.login + '/' + repo.name + ' (will fetch up to ' + repo.milestones.totalCount + ' milestones)');
            let fetchedIssues = await fetchData.load(repo, recentMilestone);
            cli_ux_1.default.action.stop(' done');
            //D - Break down the issues response in multiple batches
            const esPayloadChunked = await chunkArray_1.default(fetchedIssues, 100);
            //E- Push the results back to Elastic Search
            for (const [idx, esPayloadChunk] of esPayloadChunked.entries()) {
                cli_ux_1.default.action.start('Submitting data to ElasticSearch into ' + milestonesIndex + ' (' + (idx + 1) + ' / ' + esPayloadChunked.length + ')');
                let formattedData = '';
                for (let rec of esPayloadChunk) {
                    formattedData = formattedData + JSON.stringify({
                        index: {
                            _index: milestonesIndex,
                            _id: rec.id
                        }
                    }) + '\n' + JSON.stringify(rec) + '\n';
                }
                await client.bulk({ index: milestonesIndex, refresh: 'wait_for', body: formattedData });
                cli_ux_1.default.action.stop(' done');
            }
        }
    }
}
GhMilestones.description = 'Fetch milestones from GitHub';
GhMilestones.examples = [
    '$ github-indexer ghMilestones',
];
GhMilestones.flags = Object.assign({}, base_1.default.flags, { help: command_1.flags.help({ char: 'h' }) });
exports.default = GhMilestones;
