"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const elasticsearch_1 = require("@elastic/elasticsearch");
const command_1 = require("@oclif/command");
const cli_ux_1 = require("cli-ux");
const jsYaml = require("js-yaml");
const loadYamlFile = require("load-yaml-file");
const path = require("path");
const base_1 = require("../base");
const projects_1 = require("../schemas/projects");
const settings_1 = require("../schemas/settings");
const index_1 = require("../utils/github/fetchProjects/index");
const chunkArray_1 = require("../utils/misc/chunkArray");
class GhProjects extends base_1.default {
    /*
      The aim of this script is to fetch all projects associated with active repositories.
      It does the following:
       - Fetch a list of repositories from Elasticsearch
       - Fetch updated projects for each repository
       - Send back the content to Elasticsearch
     */
    async run() {
        const { flags } = this.parse(GhProjects);
        const userConfig = await loadYamlFile(path.join(this.config.configDir, 'config.yml'));
        const { esport, eshost, esrepo } = flags;
        const es_port = (esport !== undefined ? esport : userConfig.elasticsearch.port);
        const es_host = (eshost !== undefined ? eshost : userConfig.elasticsearch.host);
        const reposIndexName = (esrepo !== undefined ? esrepo : userConfig.elasticsearch.indices.repos);
        const indexProjectPrefix = userConfig.elasticsearch.indices.projects;
        //1- Test if an index exists, if it does not, create it.
        cli_ux_1.default.action.start('Checking if index: ' + reposIndexName + ' exists');
        const client = new elasticsearch_1.Client({ node: es_host + ':' + es_port });
        const healthCheck = await client.cluster.health();
        if (healthCheck.body.status === 'red') {
            this.log('Elasticsearch cluster is not in an healthy state, exiting');
            this.log(healthCheck.body);
            process.exit(1);
        }
        const testIndex = await client.indices.exists({ index: reposIndexName });
        if (testIndex.body === false) {
            this.error('Index: ' + reposIndexName + ' does not exists, please configure repositories first', { exit: 1 });
        }
        cli_ux_1.default.action.stop(' done');
        //2- Grab the active repositories from Elasticsearch
        cli_ux_1.default.action.start('Grabbing the active repositories from ElasticSearch');
        let esRepos = await client.search({
            index: reposIndexName,
            body: {
                from: 0,
                size: 10000,
                query: {
                    match: {
                        active: {
                            query: true
                        }
                    }
                }
            }
        });
        //console.log(esRepos.body.hits.hits);
        const activeRepos = esRepos.body.hits.hits.map(r => r._source);
        cli_ux_1.default.action.stop(' done');
        if (activeRepos.length === 0) {
            this.error('The script could not find any active repositories. Please use ghRepos and cfRepos first.', { exit: 1 });
        }
        const fetchData = new index_1.default(this.log, userConfig, this.config.configDir, cli_ux_1.default);
        const orgsFetched = [];
        this.log('Starting to grab projects');
        for (let repo of activeRepos) {
            //A - First, fetch projects at the organization level
            //A.1 - Grab org-level project (if not already fetched)
            if (!orgsFetched.includes(repo.org.login)) {
                const orgProjectsIndex = (indexProjectPrefix + repo.org.login).toLocaleLowerCase();
                const testOrgProjectIndex = await client.indices.exists({ index: orgProjectsIndex });
                if (testOrgProjectIndex.body === false) {
                    cli_ux_1.default.action.start('Elasticsearch Index ' + orgProjectsIndex + ' does not exist, creating');
                    const mappings = await jsYaml.safeLoad(projects_1.default);
                    const settings = await jsYaml.safeLoad(settings_1.default);
                    await client.indices.create({ index: orgProjectsIndex, body: { settings, mappings } });
                    cli_ux_1.default.action.stop(' created');
                }
                //A.2 - Find the most recent project
                let searchResult = await client.search({
                    index: orgProjectsIndex,
                    body: {
                        query: {
                            match_all: {}
                        },
                        size: 1,
                        sort: [
                            {
                                updatedAt: {
                                    order: 'desc'
                                }
                            }
                        ]
                    }
                });
                let recentOrgProject = null;
                if (searchResult.body.hits.hits.length > 0) {
                    recentOrgProject = searchResult.body.hits.hits[0]._source;
                }
                //A.3 - Fetch the org-level projects
                cli_ux_1.default.action.start('Grabbing org-level projects from: ' + repo.org.login);
                let fetchedOrgProjects = await fetchData.loadOrgProjects(repo.org, recentOrgProject);
                orgsFetched.push(repo.org.login);
                cli_ux_1.default.action.stop(' done');
                //A.4 - Break down the issues response in multiple batches
                const esPayloadChunked = await chunkArray_1.default(fetchedOrgProjects, 100);
                //A.5- Push the results back to Elastic Search
                for (const [idx, esPayloadChunk] of esPayloadChunked.entries()) {
                    cli_ux_1.default.action.start('Submitting data to ElasticSearch into ' + orgProjectsIndex + ' (' + (idx + 1) + ' / ' + esPayloadChunked.length + ')');
                    let formattedData = '';
                    for (let rec of esPayloadChunk) {
                        formattedData = formattedData + JSON.stringify({
                            index: {
                                _index: orgProjectsIndex,
                                _id: rec.id
                            }
                        }) + '\n' + JSON.stringify(rec) + '\n';
                    }
                    await client.bulk({ index: orgProjectsIndex, refresh: 'wait_for', body: formattedData });
                    cli_ux_1.default.action.stop(' done');
                }
            }
            // B - Second, fetch projects at the repository level
            // B.1 - Test if index exists, if not, create
            const projectsIndex = (indexProjectPrefix + repo.org.login + '_' + repo.name).toLocaleLowerCase();
            const testIndex = await client.indices.exists({ index: projectsIndex });
            if (testIndex.body === false) {
                cli_ux_1.default.action.start('Elasticsearch Index ' + projectsIndex + ' does not exist, creating');
                const mappings = await jsYaml.safeLoad(projects_1.default);
                const settings = await jsYaml.safeLoad(settings_1.default);
                await client.indices.create({ index: projectsIndex, body: { settings, mappings } });
                cli_ux_1.default.action.stop(' created');
            }
            //B.2 - Find the most recent project
            let searchResult = await client.search({
                index: projectsIndex,
                body: {
                    query: {
                        match_all: {}
                    },
                    size: 1,
                    sort: [
                        {
                            updatedAt: {
                                order: 'desc'
                            }
                        }
                    ]
                }
            });
            let recentProject = null;
            if (searchResult.body.hits.hits.length > 0) {
                recentProject = searchResult.body.hits.hits[0]._source;
            }
            //B.3 - Fetch projects from GitHub into a large array
            cli_ux_1.default.action.start('Grabbing projects for: ' + repo.org.login + '/' + repo.name + ' (will fetch up to ' + repo.projects.totalCount + ' projects)');
            let fetchedIssues = await fetchData.load(repo, recentProject);
            cli_ux_1.default.action.stop(' done');
            //B.4 - Break down the issues response in multiple batches
            const esPayloadChunked = await chunkArray_1.default(fetchedIssues, 100);
            //B.5- Push the results back to Elastic Search
            for (const [idx, esPayloadChunk] of esPayloadChunked.entries()) {
                cli_ux_1.default.action.start('Submitting data to ElasticSearch into ' + projectsIndex + ' (' + (idx + 1) + ' / ' + esPayloadChunked.length + ')');
                let formattedData = '';
                for (let rec of esPayloadChunk) {
                    formattedData = formattedData + JSON.stringify({
                        index: {
                            _index: projectsIndex,
                            _id: rec.id
                        }
                    }) + '\n' + JSON.stringify(rec) + '\n';
                }
                await client.bulk({ index: projectsIndex, refresh: 'wait_for', body: formattedData });
                cli_ux_1.default.action.stop(' done');
            }
        }
    }
}
GhProjects.description = 'Fetch projects from GitHub';
GhProjects.examples = [
    '$ github-indexer ghIssues',
];
GhProjects.flags = Object.assign({}, base_1.default.flags, { help: command_1.flags.help({ char: 'h' }) });
exports.default = GhProjects;
