"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const elasticsearch_1 = require("@elastic/elasticsearch");
const command_1 = require("@oclif/command");
const cli_ux_1 = require("cli-ux");
const fs = require("fs");
const jsYaml = require("js-yaml");
const loadYamlFile = require("load-yaml-file");
const _ = require("lodash");
const path = require("path");
const base_1 = require("../base");
const repositories_1 = require("../schemas/repositories");
const settings_1 = require("../schemas/settings");
const index_1 = require("../utils/github/fetchAffiliated/index");
const index_2 = require("../utils/github/fetchOrg/index");
const index_3 = require("../utils/github/fetchRepo/index");
const chunkArray_1 = require("../utils/misc/chunkArray");
class GhRepos extends base_1.default {
    async run() {
        const { flags } = this.parse(GhRepos);
        const userConfig = await loadYamlFile(path.join(this.config.configDir, 'config.yml'));
        const { grab, org, repo, force, esport, eshost, esrepo } = flags;
        const es_port = (esport !== undefined ? esport : userConfig.elasticsearch.port);
        const es_host = (eshost !== undefined ? eshost : userConfig.elasticsearch.host);
        const reposIndexName = (esrepo !== undefined ? esrepo : userConfig.elasticsearch.indices.repos);
        //1- Grab the repositories from GitHub
        let fetchedRepos = [];
        if (grab === 'affiliated') {
            this.log('Starting to fetch data from affiliated organizations');
            const fetchData = new index_1.default(this.log, this.error, userConfig, cli_ux_1.default);
            fetchedRepos = await fetchData.load();
        }
        else if (grab === 'org' && org !== undefined) {
            this.log('Starting to fetch data from org: ' + org);
            const fetchData = new index_2.default(this.log, userConfig, cli_ux_1.default);
            fetchedRepos = await fetchData.load(org);
        }
        else if (grab === 'repo' && org !== undefined && repo !== undefined) {
            this.log('Starting to fetch data from repo: ' + org + '/' + repo);
            const fetchData = new index_3.default(this.log, userConfig, cli_ux_1.default);
            fetchedRepos = await fetchData.load(org, repo);
        }
        //2- Test if an index exists, if it does not, create it.
        cli_ux_1.default.action.start('Checking if index: ' + reposIndexName + ' exists');
        const client = new elasticsearch_1.Client({ node: es_host + ':' + es_port });
        const healthCheck = await client.cluster.health();
        if (healthCheck.body.status === 'red') {
            this.log('Elasticsearch cluster is not in an healthy state, exiting');
            this.log(healthCheck.body);
            process.exit(1);
        }
        const testIndex = await client.indices.exists({ index: reposIndexName });
        if (testIndex.body === false) {
            cli_ux_1.default.action.start('Elasticsearch Index gh_repos does not exist, creating');
            const mappings = await jsYaml.safeLoad(repositories_1.default);
            const settings = await jsYaml.safeLoad(settings_1.default);
            //const mappings = await loadYamlFile(__dirname + '../schemas/repositories.yml')
            //const settings = await loadYamlFile(__dirname + '../schemas/settings.yml')
            await client.indices.create({ index: reposIndexName, body: { settings, mappings } });
        }
        cli_ux_1.default.action.stop(' done');
        cli_ux_1.default.action.start('Grabbing data from ElasticSearch and merging with new data');
        //3- Grab the repositories data from ElasticSearch
        let esRepos = await client.search({
            index: reposIndexName,
            body: {
                from: 0,
                size: 10000,
                query: {
                    match_all: {}
                }
            }
        });
        //4- Loop through the newly grabbed data and see if there is a corresponding result in ES
        let esPayload = [];
        fetchedRepos.map((repo) => {
            const existingRepo = _.find(esRepos.body.hits.hits, function (o) { return o._source.id === repo.id; });
            const updatedRepo = Object.assign({}, repo);
            if (existingRepo !== undefined) {
                //If the repo exist, we are only looking for the active flag
                updatedRepo.active = existingRepo._source.active;
            }
            else {
                updatedRepo.active = false;
            }
            if (force === true) {
                updatedRepo.active = true;
            }
            //If submitting only one repository, assumption is that it should be active.
            if (grab === 'repo') {
                this.log('Activating repository: ' + repo.org.login + '/' + repo.name);
                updatedRepo.active = true;
            }
            esPayload.push(updatedRepo);
        });
        cli_ux_1.default.action.stop(' done');
        this.log('About to submit (create or update) data about ' + esPayload.length + ' repo(s) to Elasticsearch');
        //Split the array in chunks of 100
        const esPayloadChunked = await chunkArray_1.default(esPayload, 100);
        //5- Push the results back to Elastic Search
        for (const [idx, esPayloadChunk] of esPayloadChunked.entries()) {
            cli_ux_1.default.action.start('Submitting data to ElasticSearch (' + (idx + 1) + ' / ' + esPayloadChunked.length + ')');
            let formattedData = '';
            for (let rec of esPayloadChunk) {
                formattedData = formattedData + JSON.stringify({
                    index: {
                        _index: reposIndexName,
                        _id: rec.id
                    }
                }) + '\n' + JSON.stringify(rec) + '\n';
            }
            await client.bulk({ index: reposIndexName, refresh: 'wait_for', body: formattedData });
            cli_ux_1.default.action.stop(' done');
        }
        //6- Update the configuration by re-downloading all data from ElasticSearch to create the configuration file
        cli_ux_1.default.action.start('Refreshing the repositories configuration file');
        esRepos = await client.search({
            index: reposIndexName,
            body: {
                from: 0,
                size: 10000,
                query: {
                    match_all: {}
                }
            }
        });
        //const esResults = _.sortBy(esRepos.body.hits.hits, [function (o) { if (o.org !== null) { return o.user } else { return '' } }])
        const esResults = _.sortBy(esRepos.body.hits.hits.map(r => r._source), ['org.login', 'name']);
        this.log('');
        this.log('All available repositories:');
        cli_ux_1.default.table(esResults, {
            name: {
                get: row => (row.org.login + '/' + row.name)
            },
            active: {
                get: row => row.active
            },
        }, {
            printLine: this.log,
        });
        this.log('');
        const configArray = esResults.map(repo => {
            return {
                [repo.org.login + '/' + repo.name]: repo.active
            };
        });
        fs.writeFileSync(path.join(this.config.configDir, 'repositories.yml'), jsYaml.safeDump(configArray));
        cli_ux_1.default.action.stop(' done');
        this.log('You can enable/disable repositories in: ' + path.join(this.config.configDir, 'repositories.yml'));
    }
}
GhRepos.description = 'Fetch repositories from GitHub (FIRST STEP, start HERE)';
GhRepos.examples = [
    '$ github-indexer ghRepo -g affiliated',
    '$ github-indexer ghRepo -g org -o jetbrains',
    '$ github-indexer ghRepo -g repo -o microsoft -r vscode',
];
GhRepos.flags = Object.assign({}, base_1.default.flags, { help: command_1.flags.help({ char: 'h' }), grab: command_1.flags.string({
        char: 'g',
        required: true,
        options: ['affiliated', 'org', 'repo'],
        description: 'Select how to fetch repositories'
    }), org: command_1.flags.string({
        char: 'o',
        required: false,
        description: 'GitHub organization login'
    }), repo: command_1.flags.string({
        char: 'r',
        required: false,
        description: 'GitHub repository name'
    }), 
    // flag with no value (-f, --force)
    force: command_1.flags.boolean({ char: 'f', default: false, description: 'Make all fetched repositories active by default' }) });
exports.default = GhRepos;
