"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const apollo_cache_inmemory_1 = require("apollo-cache-inmemory");
const apollo_client_1 = require("apollo-client");
const apollo_link_1 = require("apollo-link");
const apollo_link_http_1 = require("apollo-link-http");
const cli_ux_1 = require("cli-ux");
//import {readFileSync} from 'fs'
const _ = require("lodash");
const node_fetch_1 = require("node-fetch");
const getOrgs_1 = require("../graphql/getOrgs");
const getRepos_1 = require("../graphql/getRepos");
const getUserRepos_1 = require("../graphql/getUserRepos");
const calculateQueryIncrement_1 = require("../utils/calculateQueryIncrement");
const graphqlQuery_1 = require("../utils/graphqlQuery");
class FetchAffiliated {
    constructor(log, error, userConfig, cli) {
        this.githubToken = userConfig.github.token;
        this.githubLogin = userConfig.github.login;
        this.maxQueryIncrement = parseInt(userConfig.fetch.max_nodes, 10);
        this.log = log;
        this.error = error;
        this.cli = cli;
        this.fetchedRepos = [];
        this.githubOrgs = [];
        this.totalReposCount = 0;
        this.orgReposCount = {};
        this.errorRetry = 0;
        //this.getOrgs = readFileSync(__dirname + '/../graphql/getOrgs.graphql', 'utf8')
        //this.getRepos = readFileSync(__dirname + '/../graphql/getRepos.graphql', 'utf8')
        //this.getUserRepos = readFileSync(__dirname + '/../graphql/getUserRepos.graphql', 'utf8')
        this.getOrgs = getOrgs_1.default;
        this.getRepos = getRepos_1.default;
        this.getUserRepos = getUserRepos_1.default;
        this.rateLimit = {
            limit: 5000,
            cost: 1,
            remaining: 5000,
            resetAt: null
        };
        const httpLink = new apollo_link_http_1.HttpLink({ uri: 'https://api.github.com/graphql', fetch: node_fetch_1.default });
        const cache = new apollo_cache_inmemory_1.InMemoryCache();
        //const cache = new InMemoryCache().restore(window.__APOLLO_STATE__)
        const authMiddleware = new apollo_link_1.ApolloLink((operation, forward) => {
            // add the authorization to the headers
            operation.setContext({
                headers: {
                    authorization: this.githubToken ? `Bearer ${this.githubToken}` : '',
                }
            });
            return forward(operation).map((response) => {
                if (response.errors !== undefined && response.errors.length > 0) {
                    response.data.errors = response.errors;
                }
                return response;
            });
        });
        this.client = new apollo_client_1.default({
            link: apollo_link_1.concat(authMiddleware, httpLink),
            //link: authLink.concat(link),
            cache,
        });
    }
    async load() {
        this.log('Started load');
        cli_ux_1.default.action.start('Searching for affiliated organizations');
        await this.getOrgsPagination(null, 5);
        cli_ux_1.default.action.stop(this.githubOrgs.length + ' found');
        this.log('Initiate Organizations Repositories load');
        for (let orgObj of this.githubOrgs) {
            cli_ux_1.default.action.start('Loading repositories for organizations: ' + orgObj.login);
            await this.getReposPagination(null, 5, orgObj, 'org');
            cli_ux_1.default.action.stop(' completed');
        }
        this.log('Initiate Users own Repositories load');
        await this.getReposPagination(null, 20, { id: '', __typename: 'User', login: this.githubLogin }, 'user');
        this.log('Organizations Repositories loaded: ' + this.totalReposCount);
        return this.fetchedRepos;
    }
    sleep(ms) {
        //https://github.com/Microsoft/tslint-microsoft-contrib/issues/355
        // tslint:disable-next-line no-string-based-set-timeout
        return new Promise(resolve => setTimeout(resolve, ms));
    }
    async getOrgsPagination(cursor, increment) {
        const data = await graphqlQuery_1.default(this.client, this.getOrgs, { repo_cursor: cursor, increment }, this.rateLimit, this.log);
        //this.log('GraphQL Response:', data)
        if (data.data !== undefined) {
            if (data.data.rateLimit !== undefined) {
                this.rateLimit = data.data.rateLimit;
            }
            //updateChip(data.data.rateLimit)
            let lastCursor = await this.loadOrganizations(data);
            let queryIncrement = calculateQueryIncrement_1.default(this.githubOrgs.length, data.data.viewer.organizations.totalCount, this.maxQueryIncrement);
            if (queryIncrement > 0) {
                await this.getOrgsPagination(lastCursor, queryIncrement);
            }
        }
        else {
            this.log(data);
            this.error('Unable to make the GraphQL call to GitHub', { exit: 1 });
        }
    }
    async loadOrganizations(data) {
        let lastCursor = null;
        for (let currentOrg of data.data.viewer.organizations.edges) {
            if (currentOrg.node !== null) {
                this.githubOrgs.push(currentOrg.node);
                lastCursor = currentOrg.cursor;
            }
        }
        return lastCursor;
    }
    async getReposPagination(cursor, increment, OrgObj, type) {
        if (this.errorRetry <= 3) {
            let data = {};
            let repositories = {};
            await this.sleep(1000); // Wait 2s between requests to avoid hitting GitHub API rate limit => https://developer.github.com/v3/guides/best-practices-for-integrators/
            try {
                if (type === 'org') {
                    data = await graphqlQuery_1.default(this.client, this.getRepos, { repo_cursor: cursor, increment, org_name: OrgObj.login }, this.rateLimit, this.log);
                    repositories = data.data.viewer.organization.repositories;
                }
                else {
                    data = await graphqlQuery_1.default(this.client, this.getUserRepos, { repo_cursor: cursor, increment, login: OrgObj.login }, this.rateLimit, this.log);
                    OrgObj = data.data.user;
                    repositories = data.data.viewer.repositories;
                }
            }
            catch (error) {
                this.log(error);
            }
            //        this.log(data)
            //        this.log(OrgObj)
            if (data.data !== undefined && data.data !== null) {
                this.errorRetry = 0;
                if (this.orgReposCount[OrgObj.id] === undefined) {
                    this.orgReposCount[OrgObj.id] = 0;
                }
                if (data.data.rateLimit !== undefined) {
                    this.rateLimit = data.data.rateLimit;
                }
                //updateChip(data.data.rateLimit)
                let lastCursor = await this.loadRepositories(repositories, OrgObj);
                let queryIncrement = calculateQueryIncrement_1.default(this.orgReposCount[OrgObj.id], repositories.totalCount, this.maxQueryIncrement);
                this.log('Org: ' + OrgObj.login + ' -> Fetched Count / Remote Count / Query Increment: ' + this.orgReposCount[OrgObj.id] + ' / ' + repositories.totalCount + ' / ' + queryIncrement);
                if (queryIncrement > 0) {
                    await this.getReposPagination(lastCursor, queryIncrement, OrgObj, type);
                }
            }
            else {
                this.errorRetry = this.errorRetry + 1;
                this.log('Error loading content, current count: ' + this.errorRetry);
                await this.getReposPagination(cursor, increment, OrgObj, type);
            }
        }
        else {
            this.log('Got too many load errors, stopping');
            process.exit(1);
        }
    }
    async loadRepositories(repositories, OrgObj) {
        //    this.log('Loading from ' + OrgObj.login + ' organization')
        let lastCursor = null;
        for (let currentRepo of repositories.edges) {
            let repoObj = JSON.parse(JSON.stringify(currentRepo.node)); //TODO - Replace this with something better to copy object ?
            repoObj.org = OrgObj;
            const existingRepo = _.find(this.fetchedRepos, function (o) { return o.id === repoObj.id; });
            //There are occurences where duplicate repos might be fetched (from the organizations, then from the user).
            //Skipping if coming from the user, giving higher priority to Organixation
            if (existingRepo !== undefined && (existingRepo.org.__typename === 'Organization' && OrgObj.__typename === 'User')) {
                this.log('This repo already exists as part of an organization, skipping...');
            }
            else {
                this.fetchedRepos.push(repoObj);
            }
            lastCursor = currentRepo.cursor;
        }
        this.orgReposCount[OrgObj.id] = this.orgReposCount[OrgObj.id] + Object.entries(repositories.edges).length;
        this.totalReposCount = this.totalReposCount + Object.entries(repositories.edges).length;
        //this.log('Fetched a total of ' + this.totalReposCount + ' repositories')
        return lastCursor;
    }
}
exports.default = FetchAffiliated;
