"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const apollo_cache_inmemory_1 = require("apollo-cache-inmemory");
const apollo_client_1 = require("apollo-client");
const apollo_link_1 = require("apollo-link");
const apollo_link_http_1 = require("apollo-link-http");
const date_fns_1 = require("date-fns");
const fs_1 = require("fs");
//import { fetch } from 'apollo-env'
const node_fetch_1 = require("node-fetch");
const path = require("path");
const perf_hooks_1 = require("perf_hooks");
const getLabels_1 = require("../graphql/getLabels");
const calculateQueryIncrement_1 = require("../utils/calculateQueryIncrement");
const graphqlQuery_1 = require("../utils/graphqlQuery");
class FetchLabels {
    constructor(log, userConfig, configDir, cli) {
        this.githubToken = userConfig.github.token;
        this.githubLogin = userConfig.github.login;
        this.maxQueryIncrement = parseInt(userConfig.fetch.max_nodes, 10);
        this.configDir = configDir;
        this.log = log;
        this.cli = cli;
        this.fetchedLabels = [];
        this.errorRetry = 0;
        //this.getLabels = readFileSync(__dirname + '/../graphql/getLabels.graphql', 'utf8')
        this.getLabels = getLabels_1.default;
        this.rateLimit = {
            limit: 5000,
            cost: 1,
            remaining: 5000,
            resetAt: null
        };
        const httpLink = new apollo_link_http_1.HttpLink({ uri: 'https://api.github.com/graphql', fetch: node_fetch_1.default });
        const cache = new apollo_cache_inmemory_1.InMemoryCache();
        //const cache = new InMemoryCache().restore(window.__APOLLO_STATE__)
        const authMiddleware = new apollo_link_1.ApolloLink((operation, forward) => {
            // add the authorization to the headers
            operation.setContext({
                headers: {
                    authorization: this.githubToken ? `Bearer ${this.githubToken}` : '',
                }
            });
            return forward(operation).map((response) => {
                if (response.errors !== undefined && response.errors.length > 0) {
                    response.data.errors = response.errors;
                }
                return response;
            });
        });
        this.client = new apollo_client_1.default({
            link: apollo_link_1.concat(authMiddleware, httpLink),
            //link: authLink.concat(link),
            cache,
        });
    }
    async load(repo) {
        this.fetchedLabels = [];
        //Create stream for writing issues to cache
        this.cacheStream = fs_1.createWriteStream(path.join(this.configDir + '/cache/', 'gh_labels_' + repo.org.login + '_' + repo.name + '.ndjson'), { flags: 'a' });
        await this.getLabelsPagination(null, 5, repo);
        this.cacheStream.end();
        return this.fetchedLabels;
    }
    sleep(ms) {
        //https://github.com/Microsoft/tslint-microsoft-contrib/issues/355
        // tslint:disable-next-line no-string-based-set-timeout
        return new Promise(resolve => setTimeout(resolve, ms));
    }
    async getLabelsPagination(cursor, increment, repoObj) {
        if (this.errorRetry <= 3) {
            let data = {};
            await this.sleep(1000); // Wait 1s between requests to avoid hitting GitHub API rate limit => https://developer.github.com/v3/guides/best-practices-for-integrators/
            const t0 = perf_hooks_1.performance.now();
            try {
                data = await graphqlQuery_1.default(this.client, this.getLabels, { repo_cursor: cursor, increment, org_name: repoObj.org.login, repo_name: repoObj.name }, this.rateLimit, this.log);
            }
            catch (error) {
                this.log(error);
            }
            const t1 = perf_hooks_1.performance.now();
            const callDuration = t1 - t0;
            //        this.log(data)
            //        this.log(OrgObj)
            if (data.data !== undefined && data.data !== null) {
                this.errorRetry = 0;
                if (data.data.rateLimit !== undefined) {
                    this.rateLimit = data.data.rateLimit;
                }
                //updateChip(data.data.rateLimit)
                let lastCursor = await this.loadLabels(data, repoObj, callDuration);
                let queryIncrement = calculateQueryIncrement_1.default(this.fetchedLabels.length, data.data.repository.labels.totalCount, this.maxQueryIncrement);
                this.log('Repo: ' + repoObj.org.login + '/' + repoObj.name + ' -> Fetched Count / Remote Count / Query Increment: ' + this.fetchedLabels.length + ' / ' + data.data.repository.labels.totalCount + ' / ' + queryIncrement);
                if (queryIncrement > 0 && lastCursor !== null) {
                    await this.getLabelsPagination(lastCursor, queryIncrement, repoObj);
                }
            }
            else {
                this.errorRetry = this.errorRetry + 1;
                this.log('Error loading content, current count: ' + this.errorRetry);
                await this.getLabelsPagination(cursor, increment, repoObj);
            }
        }
        else {
            this.log('Got too many load errors, stopping');
            process.exit(1);
        }
    }
    async loadLabels(data, repoObj, callDuration) {
        //    this.log('Loading from ' + OrgObj.login + ' organization')
        let lastCursor = null;
        if (data.data.repository.labels.edges.length > 0) {
            const apiPerf = Math.round(data.data.repository.labels.edges.length / (callDuration / 1000));
            this.log('Latest call contained ' + data.data.repository.labels.edges.length + ' issues, oldest: ' + date_fns_1.format(date_fns_1.parseISO(data.data.repository.labels.edges[0].node.updatedAt), 'LLL do yyyy') + ' download rate: ' + apiPerf + ' labels/s');
        }
        for (let currentLabel of data.data.repository.labels.edges) {
            let labelObj = JSON.parse(JSON.stringify(currentLabel.node)); //TODO - Replace this with something better to copy object ?
            labelObj.repo = repoObj;
            labelObj.org = repoObj.org;
            this.fetchedLabels.push(labelObj);
            //Write the content to the cache file
            this.cacheStream.write(JSON.stringify(labelObj) + '\n');
            lastCursor = currentLabel.cursor;
        }
        return lastCursor;
    }
}
exports.default = FetchLabels;
