"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const apollo_cache_inmemory_1 = require("apollo-cache-inmemory");
const apollo_client_1 = require("apollo-client");
const apollo_link_1 = require("apollo-link");
const apollo_link_http_1 = require("apollo-link-http");
const cli_ux_1 = require("cli-ux");
//import {readFileSync} from 'fs'
const node_fetch_1 = require("node-fetch");
const getRepos_1 = require("../graphql/getRepos");
const getReposExternal_1 = require("../graphql/getReposExternal");
const getUserRepos_1 = require("../graphql/getUserRepos");
const calculateQueryIncrement_1 = require("../utils/calculateQueryIncrement");
const graphqlQuery_1 = require("../utils/graphqlQuery");
class FetchOrg {
    constructor(log, userConfig, cli) {
        this.githubToken = userConfig.github.token;
        this.githubLogin = userConfig.github.login;
        this.maxQueryIncrement = parseInt(userConfig.fetch.max_nodes, 10);
        this.log = log;
        this.cli = cli;
        this.fetchedRepos = [];
        this.totalReposCount = 0;
        this.errorRetry = 0;
        //this.getReposExternal = readFileSync(__dirname + '/../graphql/getReposExternal.graphql', 'utf8')
        //this.getRepos = readFileSync(__dirname + '/../graphql/getRepos.graphql', 'utf8')
        //this.getUserRepos = readFileSync(__dirname + '/../graphql/getUserRepos.graphql', 'utf8')
        this.getReposExternal = getReposExternal_1.default;
        this.getUserRepos = getUserRepos_1.default;
        this.getRepos = getRepos_1.default;
        this.rateLimit = {
            limit: 5000,
            cost: 1,
            remaining: 5000,
            resetAt: null
        };
        const httpLink = new apollo_link_http_1.HttpLink({ uri: 'https://api.github.com/graphql', fetch: node_fetch_1.default });
        const cache = new apollo_cache_inmemory_1.InMemoryCache();
        //const cache = new InMemoryCache().restore(window.__APOLLO_STATE__)
        const authMiddleware = new apollo_link_1.ApolloLink((operation, forward) => {
            // add the authorization to the headers
            operation.setContext({
                headers: {
                    authorization: this.githubToken ? `Bearer ${this.githubToken}` : '',
                }
            });
            return forward(operation).map((response) => {
                if (response.errors !== undefined && response.errors.length > 0) {
                    response.data.errors = response.errors;
                }
                return response;
            });
        });
        this.client = new apollo_client_1.default({
            link: apollo_link_1.concat(authMiddleware, httpLink),
            //link: authLink.concat(link),
            cache,
        });
    }
    async load(login) {
        this.log('Started load');
        cli_ux_1.default.action.start('Loading repositories for organization: ' + login);
        await this.getReposPagination(null, 10, login, {});
        cli_ux_1.default.action.stop(' completed');
        return this.fetchedRepos;
    }
    sleep(ms) {
        //https://github.com/Microsoft/tslint-microsoft-contrib/issues/355
        // tslint:disable-next-line no-string-based-set-timeout
        return new Promise(resolve => setTimeout(resolve, ms));
    }
    async getReposPagination(cursor, increment, orgLogin, orgObj) {
        if (this.errorRetry <= 3) {
            let data = {};
            await this.sleep(1000); // Wait 1s between requests to avoid hitting GitHub API rate limit => https://developer.github.com/v3/guides/best-practices-for-integrators/
            try {
                data = await graphqlQuery_1.default(this.client, this.getReposExternal, { repo_cursor: cursor, increment, org_name: orgLogin }, this.rateLimit, this.log);
            }
            catch (error) {
                this.log(error);
            }
            //        this.log(data)
            //        this.log(orgObj)
            if (data.data !== undefined && data.data !== null) {
                this.errorRetry = 0;
                if (data.data.rateLimit !== undefined) {
                    this.rateLimit = data.data.rateLimit;
                }
                const orgObj = {
                    login: data.data.organization.login,
                    name: data.data.organization.name,
                    id: data.data.organization.id,
                    url: data.data.organization.url,
                };
                let lastCursor = await this.loadRepositories(data, orgObj);
                let queryIncrement = calculateQueryIncrement_1.default(this.totalReposCount, data.data.organization.repositories.totalCount, this.maxQueryIncrement);
                this.log('Org: ' + orgObj.login + ' -> Fetched Count / Remote Count / Query Increment: ' + this.totalReposCount + ' / ' + data.data.organization.repositories.totalCount + ' / ' + queryIncrement);
                if (queryIncrement > 0) {
                    await this.getReposPagination(lastCursor, queryIncrement, orgLogin, orgObj);
                }
            }
            else {
                this.errorRetry = this.errorRetry + 1;
                this.log('Error loading content, current count: ' + this.errorRetry);
                await this.getReposPagination(cursor, increment, orgLogin, orgObj);
            }
        }
        else {
            this.log('Got too many load errors, stopping');
            process.exit(1);
        }
    }
    async loadRepositories(data, orgObj) {
        let lastCursor = null;
        for (let currentRepo of data.data.organization.repositories.edges) {
            let repoObj = JSON.parse(JSON.stringify(currentRepo.node)); //TODO - Replace this with something better to copy object ?
            repoObj.org = orgObj;
            this.fetchedRepos.push(repoObj);
            lastCursor = currentRepo.cursor;
        }
        this.totalReposCount = this.totalReposCount + Object.entries(data.data.organization.repositories.edges).length;
        //this.log('Fetched a total of ' + this.totalReposCount + ' repositories')
        return lastCursor;
    }
}
exports.default = FetchOrg;
