"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const apollo_cache_inmemory_1 = require("apollo-cache-inmemory");
const apollo_client_1 = require("apollo-client");
const apollo_link_1 = require("apollo-link");
const apollo_link_http_1 = require("apollo-link-http");
const date_fns_1 = require("date-fns");
const fs_1 = require("fs");
//import { fetch } from 'apollo-env'
const node_fetch_1 = require("node-fetch");
const path = require("path");
const perf_hooks_1 = require("perf_hooks");
const getPullrequests_1 = require("../graphql/getPullrequests");
const calculateQueryIncrement_1 = require("../utils/calculateQueryIncrement");
const graphqlQuery_1 = require("../utils/graphqlQuery");
class FetchPullrequests {
    constructor(log, userConfig, configDir, cli) {
        this.githubToken = userConfig.github.token;
        this.githubLogin = userConfig.github.login;
        this.maxQueryIncrement = parseInt(userConfig.fetch.max_nodes, 10);
        this.configDir = configDir;
        this.log = log;
        this.cli = cli;
        this.fetchedPullrequests = [];
        this.errorRetry = 0;
        //this.getPullrequests = readFileSync(__dirname + '/../graphql/getPullrequests.graphql', 'utf8')
        this.getPullrequests = getPullrequests_1.default;
        this.rateLimit = {
            limit: 5000,
            cost: 1,
            remaining: 5000,
            resetAt: null
        };
        const httpLink = new apollo_link_http_1.HttpLink({ uri: 'https://api.github.com/graphql', fetch: node_fetch_1.default });
        const cache = new apollo_cache_inmemory_1.InMemoryCache();
        //const cache = new InMemoryCache().restore(window.__APOLLO_STATE__)
        const authMiddleware = new apollo_link_1.ApolloLink((operation, forward) => {
            // add the authorization to the headers
            operation.setContext({
                headers: {
                    authorization: this.githubToken ? `Bearer ${this.githubToken}` : '',
                }
            });
            return forward(operation).map((response) => {
                if (response.errors !== undefined && response.errors.length > 0) {
                    response.data.errors = response.errors;
                }
                return response;
            });
        });
        this.client = new apollo_client_1.default({
            link: apollo_link_1.concat(authMiddleware, httpLink),
            //link: authLink.concat(link),
            cache,
        });
    }
    async load(repo, recentPullrequest) {
        this.fetchedPullrequests = [];
        //Create stream for writing PRs to cache
        this.cacheStream = fs_1.createWriteStream(path.join(this.configDir + '/cache/', repo.org.login + '_' + repo.name + '.ndjson'), { flags: 'a' });
        await this.getPullrequestsPagination(null, 5, repo, recentPullrequest);
        this.cacheStream.end();
        return this.fetchedPullrequests;
    }
    sleep(ms) {
        //https://github.com/Microsoft/tslint-microsoft-contrib/issues/355
        // tslint:disable-next-line no-string-based-set-timeout
        return new Promise(resolve => setTimeout(resolve, ms));
    }
    async getPullrequestsPagination(cursor, increment, repoObj, recentPullrequest) {
        if (this.errorRetry <= 3) {
            let data = {};
            await this.sleep(1000); // Wait 1s between requests to avoid hitting GitHub API rate limit => https://developer.github.com/v3/guides/best-practices-for-integrators/
            const t0 = perf_hooks_1.performance.now();
            try {
                data = await graphqlQuery_1.default(this.client, this.getPullrequests, { repo_cursor: cursor, increment, org_name: repoObj.org.login, repo_name: repoObj.name }, this.rateLimit, this.log);
            }
            catch (error) {
                this.log(error);
            }
            const t1 = perf_hooks_1.performance.now();
            const callDuration = t1 - t0;
            //        this.log(data)
            //        this.log(OrgObj)
            if (data.data !== undefined && data.data !== null) {
                this.errorRetry = 0;
                if (data.data.rateLimit !== undefined) {
                    this.rateLimit = data.data.rateLimit;
                }
                //updateChip(data.data.rateLimit)
                let lastCursor = await this.loadPullrequests(data, repoObj, callDuration, recentPullrequest);
                let queryIncrement = calculateQueryIncrement_1.default(this.fetchedPullrequests.length, data.data.repository.pullRequests.totalCount, this.maxQueryIncrement);
                this.log('Repo: ' + repoObj.org.login + '/' + repoObj.name + ' -> Fetched Count / Remote Count / Query Increment: ' + this.fetchedPullrequests.length + ' / ' + data.data.repository.pullRequests.totalCount + ' / ' + queryIncrement);
                if (queryIncrement > 0 && lastCursor !== null) {
                    await this.getPullrequestsPagination(lastCursor, queryIncrement, repoObj, recentPullrequest);
                }
            }
            else {
                this.errorRetry = this.errorRetry + 1;
                this.log('Error loading content, current count: ' + this.errorRetry, recentPullrequest);
                await this.getPullrequestsPagination(cursor, increment, repoObj, recentPullrequest);
            }
        }
        else {
            this.log('Got too many load errors, stopping');
            process.exit(1);
        }
    }
    async loadPullrequests(data, repoObj, callDuration, recentPullrequest) {
        //    this.log('Loading from ' + OrgObj.login + ' organization')
        let lastCursor = null;
        let stopLoad = false;
        if (data.data.repository.pullRequests.edges.length > 0) {
            const apiPerf = Math.round(data.data.repository.pullRequests.edges.length / (callDuration / 1000));
            this.log('Latest call contained ' + data.data.repository.pullRequests.edges.length + ' PRs, oldest: ' + date_fns_1.format(date_fns_1.parseISO(data.data.repository.pullRequests.edges[0].node.updatedAt), 'LLL do yyyy') + ' download rate: ' + apiPerf + ' PRs/s');
        }
        for (let currentPullrequest of data.data.repository.pullRequests.edges) {
            if (recentPullrequest !== null && new Date(currentPullrequest.node.updatedAt).getTime() < new Date(recentPullrequest.updatedAt).getTime()) {
                this.log('Pullrequest already loaded, stopping entire load');
                // Pullrequests are loaded from newest to oldest, when it gets to a point where updated date of a loaded PR
                // is equal to updated date of a local PR, it means there is no "new" content, but there might still be
                // PRs that were not loaded for any reason. So the system only stops loaded if totalCount remote is equal
                //  to the total number of PRs locally
                // Note Mar 21: This logic might be fine when the number of PRs is relatively small, definitely problematic for large repositories.
                // Commenting it out for now, it will not keep looking in the past if load is interrupted for some reason.
                //if (data.data.repository.issues.totalCount === cfgPullrequests.find({'repo.id': repoObj.id}).count()) {
                //    stopLoad = true;
                //}
                stopLoad = true;
            }
            else {
                let prObj = JSON.parse(JSON.stringify(currentPullrequest.node)); //TODO - Replace this with something better to copy object ?
                prObj.repo = repoObj;
                prObj.org = repoObj.org;
                this.fetchedPullrequests.push(prObj);
                //Write the content to the cache file
                this.cacheStream.write(JSON.stringify(prObj) + '\n');
                lastCursor = currentPullrequest.cursor;
            }
            lastCursor = currentPullrequest.cursor;
            if (stopLoad === true) {
                lastCursor = null;
            }
        }
        return lastCursor;
    }
}
exports.default = FetchPullrequests;
